package ch.zhaw.pm2.multichat.client;

import ch.zhaw.pm2.multichat.client.ClientConnectionHandler.State;
import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.ListChangeListener;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.TextArea;
import javafx.scene.control.TextField;
import javafx.scene.layout.Pane;
import javafx.stage.WindowEvent;

import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static ch.zhaw.pm2.multichat.client.ClientConnectionHandler.State.*;

public class ChatWindowController {
    private ClientConnectionHandler connectionHandler;
    private ClientMessageList messages;

    private final WindowCloseHandler windowCloseHandler = new WindowCloseHandler();

    @FXML private Pane rootPane;
    @FXML private TextField serverAddressField;
    @FXML private TextField serverPortField;
    @FXML private TextField userNameField;
    @FXML private TextField messageField;
    @FXML private TextArea messageArea;
    @FXML private Button connectButton;
    @FXML private Button sendButton;
    @FXML private TextField filterValue;


    @FXML
    public void initialize() {
        serverAddressField.setText(NetworkHandler.DEFAULT_ADDRESS.getCanonicalHostName());
        serverPortField.setText(String.valueOf(NetworkHandler.DEFAULT_PORT));
        this.messages = new ClientMessageList();

        messages.getChangedProperty().addListener(new ChangeListener<Boolean>() {
            @Override
            public void changed(ObservableValue<? extends Boolean> observable, Boolean oldValue, Boolean newValue) {
                setMessageField(messages.getMessages());
            }
        });
    }

    private void applicationClose() {
        connectionHandler.setState(DISCONNECTED);
    }

    @FXML
    private void toggleConnection () {
        if (connectionHandler == null || connectionHandler.getStateObjectProperty().get() != CONNECTED) {
            connect();
        } else {
            disconnect();
        }
    }

    private void connect() {
        try {
            messages.clear(); // clear message list
            startConnectionHandler();
            connectionHandler.connect();
        } catch(ChatProtocolException | IOException e) {
            messages.addMessage(null,null, e.getMessage(), Message.MessageType.ERROR);
        }
    }

    private void disconnect() {
        if (connectionHandler == null) {
            messages.addMessage(null,null,"No connection handler", Message.MessageType.ERROR);
            return;
        }
        try {
            connectionHandler.disconnect();
        } catch (ChatProtocolException e) {
            messages.addMessage(null,null, e.getMessage(), Message.MessageType.ERROR);
        }
    }

    @FXML
    private void message() {
        String messageString = messageField.getText().strip();
        try {
            if (connectionHandler == null) {
                messages.addMessage(null,null,"No connection handler", Message.MessageType.ERROR);
            } else if (!connectionHandler.message(messageString)) {
                messages.addMessage(null,null,"Not a valid message format.", Message.MessageType.ERROR);
            } else {
                messageField.clear();
            }
        } catch (ChatProtocolException e) {
            messages.addMessage(null,null, e.getMessage(), Message.MessageType.ERROR);
        }
    }

    //TODO: TEST
    @FXML
    private void applyFilter( ) {
        messages.getFilteredMessages(filterValue.getText().strip());
    }

    private void startConnectionHandler() throws IOException {
        String userName = userNameField.getText();
        String serverAddress = serverAddressField.getText();
        int serverPort = Integer.parseInt(serverPortField.getText());
        connectionHandler = new ClientConnectionHandler(messages,serverAddress,serverPort,userName);
        new Thread(connectionHandler).start();

        //register changelistener
        startChangeListener();

        // register window close handler
        rootPane.getScene().getWindow().addEventHandler(WindowEvent.WINDOW_CLOSE_REQUEST, windowCloseHandler);
    }

    private void terminateConnectionHandler() {
        // unregister window close handler
        rootPane.getScene().getWindow().removeEventHandler(WindowEvent.WINDOW_CLOSE_REQUEST, windowCloseHandler);
        if (connectionHandler != null) {
            connectionHandler.stopReceiving();
            connectionHandler = null;
        }
    }

    public void stateChanged(State newState) {
        // update UI (need to be run in UI thread: see Platform.runLater())
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                connectButton.setText((newState == CONNECTED || newState == CONFIRM_DISCONNECT) ? "Disconnect" : "Connect");
            }
        });
        if (newState == DISCONNECTED) {
            terminateConnectionHandler();
        }
    }

    private void startChangeListener() {
        //Listener for State
        connectionHandler.getStateObjectProperty().addListener(new ChangeListener<State>() {
            @Override
            public void changed(ObservableValue<? extends State> observable, State oldValue, State newValue) {
                stateChanged(newValue);
            }
        });

        //Listener for Username
        connectionHandler.getUserNameProperty().addListener(new ChangeListener<String>() {
            @Override
            public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
                setUserName(newValue);
            }
        });

        //Listener for Address
        connectionHandler.getServerAddressProperty().addListener(new ChangeListener<String>() {
            @Override
            public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
                setServerAddress(newValue);
            }
        });

        //Listener for Port
        connectionHandler.getServerPortProperty().addListener(new ChangeListener<Number>() {
            @Override
            public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue) {
                setServerPort(newValue.intValue());
            }
        });
    }

    public void setMessageField(String message) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                messageField.setText(message);
            }
        });
    }

    public void setUserName(String userName) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                userNameField.setText(userName);
            }
        });
    }

    public void setServerAddress(String serverAddress) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                serverAddressField.setText(serverAddress);
            }
        });
    }

    public void setServerPort(int serverPort) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                serverPortField.setText(Integer.toString(serverPort));
            }
        });
    }

    //TODO: MAKE ChangeListener
    private void redrawMessageList() {
        this.messageArea.clear();
        Platform.runLater(() -> this.messageArea.setText(messages.getFilteredMessages(filterValue.getText().strip())));
    }

    class WindowCloseHandler implements EventHandler<WindowEvent> {
        public void handle(WindowEvent event) {
            applicationClose();
        }

    }

}
